from importlib.metadata import version
from pathlib import Path

try:
    __version__ = version(Path(__file__).parent.name)
except Exception:  # noqa
    __version__ = "unknown"

try:
    with open(Path(__file__).parent.parent / ".git" / "HEAD") as g:
        head = g.read().split(":")[1].strip()
    with open(Path(__file__).parent.parent / ".git" / head) as h:
        __git_commit_hash__ = h.read().rstrip("\n")
except Exception:  # noqa
    __git_commit_hash__ = "unknown"



import pickle
from contextlib import ExitStack
from functools import wraps
from io import BytesIO, StringIO
from pathlib import Path
from re import sub
from typing import IO, Any, Callable, Hashable, Iterator, Optional, Sequence, Type

import dill
import pandas
import roifile
from bidict import bidict
from ruamel import yaml


class Pickler(dill.Pickler):
    dispatch = dill.Pickler.dispatch.copy()

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.bd_dilled = []  # [id(bidict)]
        self.bd_undilled = {}  # {id(dict): bidict}


def dill_register(t: Type) -> Callable:
    """decorator to register types to Pickler's :attr:`~Pickler.dispatch` table"""

    def proxy(func: Callable) -> Callable:
        Pickler.dispatch[t] = func
        return func

    return proxy


def undill_bidict(dct: dict, inverse: bool, undilled: dict) -> bidict:
    """ restore bidict relationships """
    bdct = undilled.get(id(dct))
    if bdct is None:
        bdct = bidict(dct)
        undilled[id(dct)] = bdct
    return bdct.inverse if inverse else bdct


@dill_register(bidict)
def dill_bidict(pickler: Pickler, bd: bidict):
    """ pickle bidict such that relationships between bidicts is preserved upon unpickling """
    if id(bd.inverse) in pickler.bd_dilled:
        pickler.save_reduce(undill_bidict, (bd.inverse._fwdm, True, pickler.bd_undilled), obj=bd)  # noqa
    else:
        pickler.bd_dilled.append(id(bd))
        pickler.save_reduce(undill_bidict, (bd._fwdm, False, pickler.bd_undilled), obj=bd)  # noqa


@dill_register(pandas.DataFrame)
def dill_dataframe(pickler: Pickler, df: pandas.DataFrame):
    """ pickle dataframe as dict to ensure compatibility """
    pickler.save_reduce(pandas.DataFrame, (df.to_dict(),), obj=df)


@wraps(pickle.dump)
def pickle_dump(obj, file: Optional[IO] = None, *args, **kwargs) -> Optional[str]:
    with ExitStack() as stack:
        if isinstance(file, (str, Path)):
            f = stack.enter_context(open(file, 'wb'))
        elif file is None:
            f = stack.enter_context(BytesIO())
        else:
            f = file
        Pickler(f, *args, **kwargs).dump(obj)
        if file is None:
            return f.getvalue()


@wraps(pickle.load)
def pickle_load(file: [bytes, str, Path, IO]) -> Any:
    if isinstance(file, bytes):
        return pickle.loads(file)
    elif isinstance(file, (str, Path)):
        with open(file, 'rb') as f:
            return pickle.load(f)
    else:
        return pickle.load(file)


class CommentedDefaultMap(yaml.CommentedMap):
    def __missing__(self, key: Hashable) -> Any:
        return None

    def __repr__(self) -> str:
        return dict.__repr__(self)


class RoundTripConstructor(yaml.RoundTripConstructor):
    pass


def construct_yaml_map(loader, node: Any) -> Iterator[CommentedDefaultMap]:
    data = CommentedDefaultMap()
    data._yaml_set_line_col(node.start_mark.line, node.start_mark.column)  # noqa
    yield data
    loader.construct_mapping(node, data, deep=True)
    loader.set_collection_style(data, node)


RoundTripConstructor.add_constructor('tag:yaml.org,2002:map', construct_yaml_map)


class RoundTripRepresenter(yaml.RoundTripRepresenter):
    pass


RoundTripRepresenter.add_representer(CommentedDefaultMap, RoundTripRepresenter.represent_dict)


@wraps(yaml.load)
def yaml_load(stream: [str, bytes, Path, IO]) -> Any:
    with ExitStack() as stack:
        y = yaml.YAML()
        y.Constructor = RoundTripConstructor
        try:
            if isinstance(stream, (str, bytes, Path)):
                stream = stack.enter_context(open(stream, 'r'))
        except (FileNotFoundError, OSError):
            pass
        return y.load(stream)


@wraps(yaml.dump)
def yaml_dump(data: Any, stream: Optional[str | bytes | Path | IO] = None, unformat: bool = False) -> Optional[str]:
    y = yaml.YAML()
    y.Representer = RoundTripRepresenter
    with StringIO() as str_io:
        y.dump(data, str_io)
        s = str_io.getvalue()
    if unformat:
        s = sub(r'<<(\w*)>>', r'{{\1}}', s)

    if stream is None:
        return s
    elif isinstance(stream, (str, bytes, Path)):
        with open(stream, 'w') as stream:
            stream.write(s)
    else:
        stream.write(s)


def get_params(parameter_file: [str, Path], template_file: [str, Path] = None,
               required: Sequence[dict] = None, ignore_empty: bool = True, replace_comments: bool = False,
               replace_values: bool = False, template_file_is_parent: bool = False,
               compare: bool = False, warn: bool = True) -> CommentedDefaultMap:
    """ Load parameters from a parameter file and parameters missing from that from the template file. Raise an error
        when parameters in required are missing. Return a dictionary with the parameters.
    """

    from .misc import cprint

    parameter_file = Path(parameter_file)
    parent_file = Path(template_file) if template_file_is_parent else parameter_file

    def yaml_load_and_format(file: Path, fmt: bool = True) -> CommentedDefaultMap:
        """ replace patterns in parameter file with parts of the parameter file path
            {{name}}: name without extension
            {{folder}}: folder
            {{suffix}}: extension
        """
        with open(file) as f:
            return yaml_load(sub(r'{{\s*(.+)\s*}}', r'{\1}' if fmt else r'<<\1>>', f.read()).format(
                name=parent_file.stem, folder=str(parent_file.parent), suffix=parent_file.suffix))

    def more_params(parameters: dict) -> None:
        """ recursively load more parameters from another file """
        more_parameters_file = parameters['more_parameters'] or parameters['more_params'] or parameters['moreParams']
        if more_parameters_file is not None:
            more_parameters_file = Path(more_parameters_file)
            if not more_parameters_file.is_absolute():
                more_parameters_file = Path(parent_file).absolute().parent / more_parameters_file
            cprint(f'<Loading more parameters from <{more_parameters_file}:.b>:g>')
            more_parameters = yaml_load_and_format(more_parameters_file)
            more_params(more_parameters)

            def add_items(sub_params, item):
                for k, v in item.items():
                    if k not in sub_params:
                        sub_params[k] = v
                    elif isinstance(v, dict):
                        add_items(sub_params[k], v)

            add_items(parameters, more_parameters)

    def check_params(parameters: dict, template: dict, path: str = '') -> None:  # noqa
        """ recursively check parameters and add defaults """
        for key, value in template.items():
            if key not in parameters and (value is not None or not ignore_empty):
                if warn:
                    cprint(f'<Parameter <{path}{key}:.b> missing, adding with value: {value}.:208>')
                parameters[key] = value
                if (isinstance(template, yaml.CommentedMap) and isinstance(parameters, yaml.CommentedMap)
                        and key in template.ca.items and isinstance(template.ca.items[key][2], yaml.CommentToken)):
                    parameters.yaml_add_eol_comment(template.ca.items[key][2].value, key)
            elif isinstance(value, dict):
                if isinstance(parameters[key], dict):
                    check_params(parameters[key], value, f'{path}{key}.')
                else:
                    if warn:
                        if parameters[key] is None:
                            cprint(f'<Parameter <{path}{key}:.b> empty, adding values: {template[key]}.:208>')
                        else:
                            cprint(f'<Overwriting <{path}{key}: {parameters[key]}:.b>.:r>')
                    parameters[key] = template[key]
            elif replace_values:
                parameters[key] = value

        if replace_comments and isinstance(template, yaml.CommentedMap) and isinstance(parameters, yaml.CommentedMap):
            # don't know how to add comments before items
            for key, value in template.ca.items.items():
                if isinstance(value[2], yaml.CommentToken):
                    parameters.yaml_add_eol_comment(value[2].value, key)

    def compare_params(parameters: Any, template: Any, reverse: bool = False, path: str = '') -> None:  # noqa
        for key, value in parameters.items():
            if isinstance(value, dict) and isinstance(template, dict) and key in template:
                compare_params(value, template[key], reverse, f'{path}{key}.')
            elif (not (value is None or isinstance(value, dict)) and
                  isinstance(template, dict) and isinstance(template.get(key), dict)):
                if reverse:
                    cprint(f'<New parameter: <{path}{key}:.b>: {value} is not a dictionary anymore.:r>')
                else:
                    cprint(f'<Old parameter: <{path}{key}:.b>: {value} is now a dictionary.:r>')
            elif template is None or isinstance(template, dict) and key not in template:
                if reverse:
                    cprint(f'<New parameter: <{path}{key}:.b>: {value}.:g>')
                else:
                    cprint(f'<Old parameter: <{path}{key}:.b>: {value}.:208>')

    def check_required(parameters: dict, required: Sequence[dict]) -> None:  # noqa
        if required is not None:
            for p in required:
                if isinstance(p, dict):
                    for key, value in p.items():
                        check_required(parameters[key], value)
                else:
                    if p not in parameters:
                        raise Exception(f'Parameter {p} not given in parameter file.')

    def check_new_lines(parameters: dict, gap: int = 2) -> None:  # noqa
        n = len(parameters) - 1
        for i, (key, value) in enumerate(parameters.items()):  # noqa
            if isinstance(parameters, yaml.CommentedMap):
                if key in parameters.ca.items and parameters.ca.items[key][2] is not None:
                    comment = parameters.ca.items[key][2].value.rstrip('\n') + '\n'
                    if (gap == 2 or (i == n and gap)) and not isinstance(value, dict):
                        comment += '\n'
                    parameters.ca.items[key][2].value = comment
                elif (gap == 2 or (i == n and gap)) and not isinstance(value, dict):
                    parameters.yaml_add_eol_comment(' ', key)
                    parameters.ca.items[key][2].value = '\n\n'
            if isinstance(value, dict):
                check_new_lines(value, gap == 2 or (i == n and gap))

    params = yaml_load_and_format(parameter_file)
    more_params(params)
    check_required(params, required)

    if template_file is not None:
        template = yaml_load_and_format(template_file, False)

        if compare:
            compare_params(template, params)
            compare_params(params, template, reverse=True)
        check_params(params, template)

    check_new_lines(params)
    return params


def save_roi(file: [str, Path], coordinates: pandas.DataFrame, shape: tuple, columns: Optional[Sequence[str]] = None,
             name: Optional[str] = None):
    if columns is None:
        columns = 'xyCzT'
    coordinates = coordinates.copy()
    if '_' in columns:
        coordinates['_'] = 0
    # if we save coordinates too close to the right and bottom of the image (<1 px) the roi won't open on the image
    if not coordinates.empty:
        coordinates = coordinates.query(f'-0.5<={columns[0]}<{shape[1]-1.5} & -0.5<={columns[1]}<{shape[0]-1.5} &'
                                        f' -0.5<={columns[3]}<={shape[3]-0.5}')
    if not coordinates.empty:
        roi = roifile.ImagejRoi.frompoints(coordinates[list(columns[:2])].to_numpy().astype(float))
        roi.roitype = roifile.ROI_TYPE.POINT
        roi.options = roifile.ROI_OPTIONS.SUB_PIXEL_RESOLUTION
        roi.counters = len(coordinates) * [0]
        roi.counter_positions = (1 + coordinates[columns[2]].to_numpy() +
                                 coordinates[columns[3]].to_numpy().round().astype(int) * shape[2] +
                                 coordinates[columns[4]].to_numpy() * shape[2] * shape[3]).astype(int)
        if name is None:
            roi.name = ''
        else:
            roi.name = name
        roi.version = 228
        roi.tofile(file)



from __future__ import annotations

from abc import ABCMeta, abstractmethod
from copy import copy
from functools import cached_property
from typing import Any, Callable, Sequence

import numpy as np
from numpy.typing import ArrayLike
from parfor import pmap
from scipy import special, stats
from scipy.optimize import OptimizeResult, minimize

Number = int | float | complex


class Fit(metaclass=ABCMeta):
    bounds = None

    def __init__(self, x: ArrayLike, y: ArrayLike,
                 w: [ArrayLike, None] = None, s: [ArrayLike, None] = None,
                 fit_window: Sequence[float] = None, log_scale: bool = False, fit_s: bool = True,
                 p0: ArrayLike = None) -> None:
        x = np.asarray(x)
        y = np.asarray(y)
        w = np.ones_like(x) if w is None else np.asarray(w)
        if log_scale:
            s = np.ones_like(x) if s is None else np.asarray(s) / np.abs(y)
        else:
            s = np.ones_like(x) if s is None else np.asarray(s)

        if fit_window:
            idx = (fit_window[0] <= x) & (x < fit_window[1])
            x, y, w, s = x[idx], y[idx], w[idx], s[idx]

        self.x, self.y, self.w, self.s = nonnan(x, y, w, s)
        self.log_scale = log_scale
        self.fit_s = fit_s
        self.n = np.sum(self.w)
        self.p_ci95 = None
        self.r_squared = None
        self.chi_squared = None
        self.r_squared_adjusted = None
        self.p0_manual = p0

    @property
    @abstractmethod
    def n_p(self) -> int:
        pass

    @property
    @abstractmethod
    def p0(self) -> ArrayLike:
        pass

    @staticmethod
    @abstractmethod
    def fun(p: ArrayLike, x: Number | ArrayLike) -> ArrayLike:
        pass

    def dfun(self, p: ArrayLike, x: ArrayLike, diffstep: float = 1e-6) -> np.ndarray:
        """ d fun / dp_i for each p_i in p, this default function will calculate it numerically """
        eps = np.spacing(1)
        deriv = np.zeros((len(p), len(x)))
        f0 = np.asarray(self.fun(p, x))
        p = np.asarray(p)
        for i in range(len(p)):
            ph = p.copy()
            ph[i] = p[i] * (1 + diffstep) + eps
            f = np.asarray(self.fun(ph, x))
            deriv[i] = (f - f0) / (ph[i] - p[i])
        return deriv

    def evaluate(self, x: Number | ArrayLike = None) -> tuple[ArrayLike, ArrayLike]:
        if x is None:
            x = np.linspace(np.nanmin(self.x), np.nanmax(self.x))
        else:
            x = np.asarray(x)
        return x.real, self.fun(self.p, x)

    def evaluate_ci(self, x: Number | ArrayLike = None) -> tuple[ArrayLike, ArrayLike, ArrayLike]:
        if x is None:
            x = np.linspace(np.nanmin(self.x), np.nanmax(self.x))
        else:
            x = np.asarray(x)
        f = self.fun(self.p, x)
        df = np.sqrt(np.sum((self.dfun(self.p, x).T * self.p_ci95).T ** 2, 0))
        return x.real, f - df, f + df

    def get_cost_fun(self) -> Callable[[ArrayLike], float]:
        s = self.s if self.fit_s else 1
        eps = np.spacing(0)
        if self.log_scale:
            def cost(p: ArrayLike) -> float:
                with np.errstate(divide='ignore'):
                    return np.nansum(np.abs(
                        self.w / s * (np.log(self.y) - np.log(np.clip(self.fun(p, self.x), eps, None))) ** 2))
        else:
            def cost(p: ArrayLike) -> float:
                with np.errstate(divide='ignore'):
                    return np.nansum(np.abs(self.w / s * (self.y - self.fun(p, self.x)) ** 2))
        return cost

    def fit(self) -> Fit:
        _ = self.r
        return self

    @cached_property
    def r(self) -> OptimizeResult:
        if not hasattr(self, 'p0_manual'):
            self.p0_manual = None
        with np.errstate(divide='ignore'):
            if len(self.x):
                r = minimize(self.get_cost_fun(), np.asarray(self.p0 if self.p0_manual is None else self.p0_manual),
                             method='Nelder-Mead', bounds=self.bounds, options=dict(maxiter=400 * self.n_p))
            else:
                r = OptimizeResult(fun=np.nan, message='Empty data', nfev=0, nit=0, status=1, success=False,
                                   x=np.full(self.n_p, np.nan))
            if self.log_scale:
                self.p_ci95, self.r_squared = fminerr(
                    lambda p, x: np.log(self.fun(p, x)), self.sort(r.x), np.log(self.y), (self.x,), self.w, self.s)
            else:
                self.p_ci95, self.r_squared = fminerr(self.fun, self.sort(r.x), self.y, (self.x,), self.w, self.s)
            if self.n - self.n_p - 1 > 0:
                self.r_squared_adjusted = 1 - (1 - self.r_squared) * (self.n - 1) / (self.n - self.n_p - 1)
            else:
                self.r_squared_adjusted = np.nan
            return r

    @staticmethod
    def sort(p: np.ndarray) -> np.ndarray:
        return p

    @property
    def p(self) -> np.ndarray:
        return np.full(self.n_p, np.nan) if self.r is None else self.sort(self.r.x)

    @property
    def log_likelihood(self) -> float:
        return -self.n * np.log(2 * np.pi * self.r.fun / (self.n - 1)) / 2 - (self.n - 1) / 2

    @property
    def bic(self) -> float:
        """ Bayesian Information Criterion: the fit with the smallest bic should be the best fit """
        return self.n_p * np.log(self.n) - 2 * self.log_likelihood

    def ftest(self, fit2) -> float:
        """ returns the p-value for the hypothesis that fit2 is the better fit,
            assuming fit2 is the fit with more free parameters
            if the fits are swapped the p-value will be negative """
        if not np.all(self.x == fit2.x):
            raise ValueError('Only two fits on the same data can be compared.')
        if self.n_p == fit2.n_p:
            raise ValueError('The two fits cannot have the same number of parameters.')
        rss1 = self.get_cost_fun()(self.p)
        rss2 = fit2.get_cost_fun()(fit2.p)
        swapped = np.argmin((self.n_p, fit2.n_p))
        if swapped and rss1 > rss2:
            return -1
        elif not swapped and rss1 < rss2:
            return 1
        else:
            n = self.n_p if swapped else fit2.n_p
            dn = np.abs(self.n_p - fit2.n_p)
            f_value = (np.abs(rss1 - rss2) / dn) / ((rss1 if swapped else rss2) / (self.n - n))
            p_value = stats.f(dn, self.n - n).sf(f_value)
            return -p_value if swapped else p_value

    def reset(self, log_scale: bool = None, fit_s: bool = True, p0: ArrayLike = None) -> Fit:
        new = copy(self)
        if log_scale is not None:
            if log_scale and not self.log_scale:
                new.s /= np.abs(new.y)
            if not log_scale and self.log_scale:
                new.s *= np.abs(new.y)
            new.log_scale = log_scale
        if fit_s is not None:
            new.fit_s = fit_s
        if p0 is not None:
            new.p0_manual = p0
        if hasattr(new, 'r'):
            delattr(new, 'r')
        return new

    def get_best_p0(self, space: np.ndarray) -> Fit:
        bounds = np.array([(-np.inf if i[0] is None else i[0], np.inf if i[1] is None else i[1])
                           for i in self.bounds]).T
        p0 = np.clip(np.stack(np.meshgrid(*self.n_p * (space,)), 0).reshape((self.n_p, -1)).T * self.p0, *bounds)
        return self.reset(p0=p0[np.argmin(pmap(self.get_cost_fun(), p0, desc='finding best p0', leave=False))])


class Exponential1(Fit):
    n_p = 2
    bounds = ((0, None), (0, None))

    @property
    def p0(self) -> ArrayLike:
        """ y = a*exp(-t/tau)
            return a, tau
        """
        x, y = finite(self.x.astype('complex'), np.log(self.y.astype('complex')))
        if len(x) < 2:
            return [1, 1]
        else:
            q = np.polyfit(x, y, 1)
            return [np.clip(value.real, *bound) for value, bound in zip((np.exp(q[1]), -1 / q[0]), self.bounds)]

    @staticmethod
    def fun(p: ArrayLike, x: Number | ArrayLike) -> ArrayLike:
        return p[0] * np.exp(-x / p[1])

    # def dfun(self, p, x, diffstep=None):
    #     e = np.exp(-x / p[1])
    #     return np.vstack((e, p[0] * x * e / p[1] ** 2))


class Exponential2(Fit):
    n_p = 4
    bounds = ((0, None), (0, 1), (0, None), (0, None))

    @property
    def p0(self) -> ArrayLike:
        """ y = A(a*exp(-t/tau_0) + (1-a)*exp(-t/tau_1))
            return A, a, tau_0, tau_1
        """
        n = len(self.x) // 2
        y0 = np.nanmax(self.y)
        q = Exponential1(self.x[n:], self.y[n:] / y0).p0
        return [np.clip(value, *bound)
                for value, bound in zip((y0, 1 - q[0], q[1] / 3, q[1]), self.bounds)]

    @staticmethod
    def fun(p: ArrayLike, x: Number | ArrayLike) -> ArrayLike:
        return p[0] * (p[1] * np.exp(-x / p[2]) + (1 - p[1]) * np.exp(-x / p[3]))

    # def dfun(self, p, x, diffstep=None):
    #     e0 = np.exp(-x / p[2])
    #     e1 = np.exp(-x / p[3])
    #     return np.vstack((p[1] * e0 + (1 - p[1]) * e1, p[0] * (e0 - e1),
    #                       p[0] * p[1] * e0 / p[2] ** 2, p[0] * (1 - p[1]) * e1 / p[3] ** 2))


class Exponential3(Fit):
    n_p = 6
    bounds = ((0, None), (0, 1), (0, 1), (0, None), (0, None), (0, None))

    @property
    def p0(self) -> ArrayLike:
        """ y = A(a*exp(-t/tau_0) + b*exp(-t/tau_1) + (1-a-b)*exp(-t/tau-2))
            return A, a, b, tau_0, tau_1, tau_2
        """
        n = len(self.x) // 2
        y0 = np.nanmax(self.y)
        q = Exponential2(self.x[n:], self.y[n:] / y0).p0
        return [np.clip(value, *bound)
                for value, bound in zip((y0, 0.3, 0.3, q[2] / 3, q[3] / 3, q[3]), self.bounds)]

    @staticmethod
    def fun(p: ArrayLike, x: Number | ArrayLike) -> ArrayLike:
        return p[0] * (p[1] * np.exp(-x / p[3]) + p[2] * np.exp(-x / p[4]) +
                       (1 - p[1] - p[2]) * np.exp(-x / p[5]))


class Powerlaw(Fit):
    n_p = 2

    @property
    def p0(self) -> ArrayLike:
        """ y = (x/tau)^alpha
            return alpha, tau
        """
        q = np.polyfit(*finite(np.log(self.x.astype('complex')), np.log(self.y.astype('complex'))), 1)
        return q[0].real, np.exp(-q[1] / q[0]).real

    @staticmethod
    def fun(p: ArrayLike, x: Number | ArrayLike) -> ArrayLike:
        return ((np.asarray(x).astype('complex') / p[1]) ** p[0]).real


class GammaCDF(Fit):
    n_p = 2

    @property
    def p0(self) -> ArrayLike:
        """ y = γ(k, x / θ) / Γ(k)
        """
        m = np.sum(-self.x[1:] * np.diff(self.y))
        v = np.sum(-(self.x[1:] - m) ** 2 * np.diff(self.y))
        return m ** 2 / v, v / m  # A, k, theta

    @staticmethod
    def fun(p: ArrayLike, x: Number | ArrayLike) -> ArrayLike:
        """ p: k, theta """
        return 1 - special.gammainc(p[0], x / p[1])


def finite(*args: ArrayLike) -> list[np.ndarray]:
    idx = np.prod([np.isfinite(arg) for arg in args], 0).astype(bool)
    return [np.asarray(arg)[idx] for arg in args]


def nonnan(*args: ArrayLike) -> list[np.ndarray]:
    idx = np.prod([~np.isnan(arg) for arg in args], 0).astype(bool)
    return [np.asarray(arg)[idx] for arg in args]


def fminerr(fun: Callable[[ArrayLike, Any], float], a: ArrayLike, y: ArrayLike,
            args: tuple[Any] = (), w: ArrayLike = None, s: ArrayLike = None,
            diffstep: float = 1e-6) -> tuple[np.ndarray, float]:
    """ Error estimation of a fit

        Inputs:
        fun:  function which was fitted to data
        a:    function parameters
        y:    ydata
        args: extra arguments to fun
        w:    weights
        s:    error on y (std)

        Outputs:
        da:    95% confidence interval
        R2:    R^2

        Example:
        x = np.array((-3,-1,2,4,5))
        a = np.array((2,-3))
        y = (15,0,5,30,50)
        fun = lambda a: a[0]*x**2+a[1]
        dp, R2 = fminerr(fun, p, y)

        adjusted from Matlab version by Thomas Schmidt, Leiden University
        wp@tl2020
    """
    eps = np.spacing(1)
    a = np.array(a).flatten()
    y = np.array(y).flatten()
    w = np.ones_like(y) if w is None else np.asarray(w).flatten()
    s = np.zeros_like(y) if s is None else np.asarray(s).flatten()

    n_data = np.size(y)
    n_par = np.size(a)

    if n_data > n_par:
        f0 = np.array(fun(a, *args)).flatten()
        var_res = (np.sum(w * (f0 - y) ** 2) + np.sum(w * s ** 2)) / (np.sum(w) - n_par)

        # calculate R^2
        ss_tot = np.sum(w * (y - np.nanmean(y)) ** 2)
        ss_res = np.sum(w * (y - f0) ** 2)
        r_squared = 1 - ss_res / ss_tot

        # calculate derivatives
        jac = np.zeros((n_data, n_par), dtype='complex')
        for i in range(n_par):
            ah = a.copy()
            ah[i] = np.clip(a[i] * (1 + diffstep), eps, None)
            f = np.array(fun(ah, *args)).flatten()
            jac[:, i] = (f - f0) / (ah[i] - a[i])

        hesse = np.matmul(jac.T, jac)

        try:
            if np.linalg.matrix_rank(hesse) == np.shape(hesse)[0]:
                da = np.sqrt(var_res * np.diag(np.linalg.inv(hesse)))
            else:
                da = np.sqrt(var_res * np.diag(np.linalg.pinv(hesse)))
        except (Exception,):
            da = np.full_like(a, np.nan)
        return 1.96 * da.real, r_squared.real
    else:
        return np.full_like(a, np.nan), np.nan



from __future__ import annotations

import os
import tempfile
from contextlib import redirect_stdout
from inspect import getfullargspec
from io import StringIO
from itertools import product
from pathlib import Path
from typing import Any, Callable, Sequence

import numpy as np
import pandas
from cellpose import models
from csbdeep.utils import normalize
from ndbioimage import Imread
from numpy.typing import ArrayLike
from scipy.interpolate import interp1d
from scipy.ndimage import distance_transform_edt
from scipy.spatial.distance import cdist
from skimage.segmentation import watershed
from tiffwrite import FrameInfo, IJTiffFile
from tqdm.auto import tqdm

from .findcells import findcells
from .pytrackmate import trackmate_peak_import

os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'
from stardist.models import StarDist2D  # noqa

try:
    import imagej
    import scyjava

    def kill_vm():
        if scyjava.jvm_started():
            scyjava.shutdown_jvm()

except ImportError:
    imagej = None
    scyjava = None

    def kill_vm():
        return


def label_dist(labels: np.ndarray, lbl: int, mask: np.ndarray = None) -> np.ndarray:
    """ make an array with distances to the edge of the lbl in labels, negative outside, positive inside """
    lbl_mask = (labels == lbl)
    dist = -distance_transform_edt(lbl_mask == 0)
    dist[dist < 0] += 1
    dist += distance_transform_edt(lbl_mask == 1)
    dist[(labels != lbl) * (labels != 0)] = -np.inf
    if mask is not None:
        dist[mask] = -np.inf
    return dist


def interp_label(t: int, ts: Sequence[int], labels: Sequence[np.ndarray], lbl: int,
                 mask: np.ndarray = None) -> np.ndarray:
    """ return a label field with lbl at time q interpolated from labels at times ts """
    return lbl * (interp1d(ts, np.dstack([label_dist(label, lbl, mask) for label in labels]),
                           fill_value=np.zeros_like(labels[0]), bounds_error=False)(t) > 0)


class SwapLabels:
    def __init__(self, tracks: pandas.DataFrame, min_frames: int = None) -> None:
        if min_frames:
            tracks = tracks.groupby('label').apply(lambda df: df.assign(n_frames=len(df))).query(
                f'n_frames > @min_frames', local_dict=dict(min_frames=min_frames)).reset_index(drop=True)
        # ensure that labels are consecutive
        if not tracks.empty:
            d = {u: i for i, u in enumerate(tracks['label'].unique(), 1)}
            tracks['label'] = tracks.apply(lambda s: d[s['label']], 1)
        self.tracks = tracks

    def __call__(self, im: Imread, frame_in: np.ndarray, c: int, z: int, t: int) -> np.ndarray:
        frame_out = np.zeros_like(frame_in)
        for i, j in self.tracks.query('t == @t', local_dict=dict(t=t))[['median_intensity', 'label']].to_numpy():
            frame_out[frame_in == i] = j
        return frame_out


def sort_labels(tracks: pandas.DataFrame) -> pandas.DataFrame:
    """ make labels consistent across different runs """
    relabel_dict = {int(key): value for value, key in
                    enumerate(tracks.groupby('label').aggregate('mean').sort_values('area').index, 1)}
    return tracks.groupby('label').apply(lambda x: x.assign(label=relabel_dict[x['label'].mean()]))


def get_time_points(t: int, missing: Sequence[int]) -> tuple[int, int]:
    t_a = t - 1
    while t_a in missing:
        t_a -= 1
    t_b = t + 1
    while t_b in missing:
        t_b += 1
    return t_a, t_b


def interpolate_missing(tracks: pandas.DataFrame, t_len: int = None) -> pandas.DataFrame:
    """ interpolate the position of the cell in missing frames """
    missing = []
    for cell in tracks['label'].unique():
        h = tracks.query('label==@cell', local_dict=dict(cell=cell))
        if t_len is None:
            t_missing = list(set(range(int(h['t'].min()), int(h['t'].max()))) - set(h['t']))
        else:
            t_missing = list(set(range(t_len)) - set(h['t']))
        g = pandas.DataFrame(np.full((len(t_missing), tracks.shape[1]), np.nan), columns=tracks.columns)
        g['t'] = t_missing
        g['t_stamp'] = t_missing
        g['x'] = np.interp(t_missing, h['t'], h['x'])
        g['y'] = np.interp(t_missing, h['t'], h['y'])
        g['label'] = cell
        missing.append(g)
    return pandas.concat(missing, ignore_index=True)


def substitute_missing(tracks: pandas.DataFrame, missing: pandas.DataFrame, distance: int = 1) -> pandas.DataFrame:
    """ relabel rows in tracks if they overlap with a row in missing """
    for _, row in missing.iterrows():
        a = tracks.query(f't==@t & (x-@x)**2 + (y-@y)**2 < @distance',
                         local_dict=dict(t=row['t'], x=row['x'], y=row['y'], distance=distance)).copy()
        a['label'] = row['label']
        if len(a) == 1:
            tracks.loc[a.index[0], 'label'] = row['label']
        elif len(a) > 1:
            idx = ((a[['x', 'y']] - row[['x', 'y']].tolist()) ** 2).sum(1).idxmin()
            tracks.loc[idx, 'label'] = row['label']
    return tracks


def filter_kwargs(function: Callable, kwargs: dict[str, Any]) -> dict[str, Any]:
    args = getfullargspec(function)
    return {key: value for key, value in kwargs.items() if key in args.args + args.kwonlyargs}


def get_xy(im: ArrayLike) -> tuple[float, float]:
    im = np.asarray(im)
    x, y = np.meshgrid(np.arange(im.shape[1]), np.arange(im.shape[0]))
    s = np.sum(im)
    return np.sum(im * y) / s, np.sum(im * x) / s


def connect_nuclei_with_cells(nuclei: ArrayLike, cells: ArrayLike) -> np.ndarray:
    nuclei = np.asarray(nuclei)
    cells = np.asarray(cells)
    i_nuclei = np.array([i for i in np.unique(nuclei) if i > 0])
    i_cells = np.array([i for i in np.unique(cells) if i > 0])
    j = (nuclei.flatten()) > 0 | (cells.flatten() > 0)
    nuclei_flat = nuclei.flatten()[j]
    cells_flat = cells.flatten()[j]
    jaccard = cdist(np.vstack([nuclei_flat == i for i in i_nuclei]).astype(int),
                    np.vstack([cells_flat == i for i in i_cells]).astype(int), 'jaccard')
    idx = np.where(jaccard < 0.95)
    idx = np.vstack((i_nuclei[idx[0]], i_cells[idx[1]])).T

    d = {}
    for n in np.unique(idx[:, 0]):
        c_i = set(idx[idx[:, 0] == n, 1])
        n_i = tuple(np.unique(idx[np.isin(idx[:, 1], tuple(c_i)), 0]))
        d[n_i] = d.get(n_i, set()) | c_i  # type: ignore

    cells_new = np.zeros_like(cells)
    x, y = np.meshgrid(range(cells.shape[1]), range(cells.shape[0]))
    for n, c in d.items():
        if len(n) == 1:
            cells_new[np.isin(cells, tuple(c))] = n
        else:
            mask = np.zeros_like(cells)
            mask[np.isin(cells, tuple(c))] = 1
            centers = [get_xy(nuclei == i) for i in n]
            dist = np.min([(x - i[1]) ** 2 + (y - i[0]) ** 2 for i in centers], 0)
            markers = np.zeros_like(cells)
            for i, n_c in zip(centers, n):
                markers[int(round(i[0])), int(round(i[1]))] = n_c
            cells_new += watershed(dist, markers=markers, mask=mask)
    cells_new[nuclei > 0] = nuclei[nuclei > 0]
    cells_new[np.isin(cells_new, np.setdiff1d(cells_new, nuclei))] = 0
    return cells_new


def trackmate_fiji(file_in: Path, file_out: Path, fiji_path: Path = None,
                   channel: int = 0, **kwargs: dict[str, [str, int, float, bool]]) -> None:
    if fiji_path is None:
        fiji_path = Path('/DATA/opt/Fiji.app')
    if fiji_path.exists():
        ij = imagej.init(str(fiji_path))
    else:
        ij = imagej.init('sc.fiji:fiji')
    settings = dict(file_in=str(file_in), file_out=str(file_out), TARGET_CHANNEL=1 + channel,
                    MIN_AREA=20, MAX_FRAME_GAP=2, ALTERNATIVE_LINKING_COST_FACTOR=1.05,
                    LINKING_MAX_DISTANCE=15.0, GAP_CLOSING_MAX_DISTANCE=15.0, SPLITTING_MAX_DISTANCE=15.0,
                    ALLOW_GAP_CLOSING=True, ALLOW_TRACK_SPLITTING=False, ALLOW_TRACK_MERGING=False,
                    MERGING_MAX_DISTANCE=15.0, CUTOFF_PERCENTILE=0.9)
    settings.update({key.upper(): value for key, value in kwargs.items() if key in settings})  # noqa
    with open(Path(__file__).parent / 'trackmate.jy') as f:
        ij.py.run_script('py', f.read(), settings)
    ij.dispose()


def trackmate(tif_file: Path, tiff_out: Path, table_out: Path = None, min_frames: int = None,
              **kwargs: dict[str, str]) -> None:
    """ run trackmate to make sure cells have the same label in all frames, relabel even if there's just one frame,
        to make sure that cell numbers are consecutive """

    with Imread(tif_file, axes='ctyx', dtype=int) as im:
        if im.shape['t'] > 1:
            xml_file = tif_file.with_suffix('.xml')
            trackmate_fiji(tif_file, xml_file, channel=im.shape['c'], **kwargs)
            tracks = trackmate_peak_import(str(xml_file), get_tracks=True)
            missing = interpolate_missing(tracks)
            tracks = substitute_missing(tracks, missing)
            tracks = sort_labels(tracks)
            missing = interpolate_missing(tracks)
            tracks = pandas.concat((tracks, missing), ignore_index=True)
        else:
            cells = np.unique(im)
            tracks = pandas.DataFrame(np.vstack((np.arange(1, len(cells)), cells[cells > 0])).T,
                                      columns=['label', 'median_intensity']).assign(t=0)
            missing = None
        if table_out:
            tracks.to_csv(table_out, sep='\t', index=False)

        # relabel the labels according to the tracks and also add missing labels by interpolation
        im.frame_decorator = SwapLabels(tracks, min_frames)
        dtype = 'uint8' if im.frame_decorator.tracks['label'].max() <= 255 else 'uint16'
        with IJTiffFile(tiff_out, (im.shape['c'], 1, im.shape['t']), pxsize=im.pxsize_um,
                        colormap='glasbey', dtype=dtype) as tif:
            for c, t in tqdm(product(range(im.shape['c']), range(im.shape['t'])),
                             total=im.shape['c'] * im.shape['t'],
                             desc='reordering cells with trackmate', leave=False):
                frame = np.asarray(im[c, t])
                if missing is not None:
                    missing_t = missing.query('t==@t', local_dict=dict(t=t))
                    for cell in missing_t['label'].unique():
                        time_points = get_time_points(t, missing.query('label==@cell',
                                                                       local_dict=dict(cell=cell))['t'].tolist())
                        a = interp_label(t, time_points, [im[c, i] for i in time_points], int(cell), frame > 0)
                        frame += a
                # TODO: transforms
                tif.save(frame, c, 0, t)


def run_stardist(image: Path, tiff_out: Path, channel_cell: int, *,
                 model_type: str = None, table_out: Path = None, tm_kwargs: dict[str, str] = None) -> None:
    if model_type is None:
        model_type = '2D_versatile_fluo'
    with redirect_stdout(StringIO()):
        model = StarDist2D.from_pretrained(model_type)
    tm_kwargs = tm_kwargs or {}

    with tempfile.TemporaryDirectory() as tempdir:
        tif_file = Path(tempdir) / 'tm.tif'

        with Imread(image, axes='ctyx') as im:
            with IJTiffFile(tif_file, (1, 1, im.shape['t']), pxsize=im.pxsize_um) as tif:
                for t in tqdm(range(im.shape['t']), total=im.shape['t'], desc='running stardist',
                              disable=im.shape['t'] < 10):
                    tif.save(model.predict_instances(normalize(im[channel_cell, t]))[0], 0, 0, t)

        trackmate(tif_file, tiff_out, table_out, **tm_kwargs)


class CellPoseTiff(IJTiffFile):
    def __init__(self, model: models.Cellpose, cp_kwargs: dict[str, str] = None, *args: Any, **kwargs: Any) -> None:
        self.model = model
        self.cp_kwargs = cp_kwargs or {}
        super().__init__(*args, **kwargs)

    def compress_frame(self, frame: tuple[ArrayLike]) -> Sequence[FrameInfo]:
        if len(frame) == 1:
            cells = self.model.eval(np.stack(frame, 0), channel_axis=0, channels=[[0, 0]],  # noqa
                                            **self.cp_kwargs)[0]
            return super().compress_frame(cells.astype(self.dtype))
        else:
            cells = self.model.eval(np.stack(frame, 0), channel_axis=0, channels=[[1, 0]],  # noqa
                                    **self.cp_kwargs)[0]
            nuclei = self.model.eval(np.stack(frame, 0), channel_axis=0, channels=[[2, 0]],  # noqa
                                     **self.cp_kwargs)[0]
            cells = connect_nuclei_with_cells(nuclei, cells)
            return [super().compress_frame(cells.astype(self.dtype))[0],
                    super().compress_frame(nuclei.astype(self.dtype))[0][:2] + ((1, 0, 0),)]


def run_cellpose(image: Path, tiff_out: Path, channel_cell: int, channel_nuc: int = None, *,
                 model_type: str = None, table_out: Path = None,
                 cp_kwargs: dict[str, str] = None, tm_kwargs: dict[str, str] = None) -> None:
    cp_kwargs = cp_kwargs or {}
    tm_kwargs = tm_kwargs or {}
    model = models.Cellpose(gpu=False, model_type=model_type)
    cp_kwargs = filter_kwargs(model.eval, cp_kwargs)

    with tempfile.TemporaryDirectory() as tempdir:
        tif_file = Path(tempdir) / 'tm.tif'

        with Imread(image, axes='ctyx') as im:
            with CellPoseTiff(model, cp_kwargs, tif_file, (1 if channel_nuc is None else 2, 1, im.shape['t']),
                              pxsize=im.pxsize_um) as tif:
                for t in tqdm(range(im.shape['t']), total=im.shape['t'], desc='running cellpose',
                              disable=im.shape['t'] < 10):
                    tif.save((im[channel_cell, t],) if channel_nuc is None else
                             (im[channel_cell, t], im[channel_nuc, t]), 0, 0, t)

        trackmate(tif_file, tiff_out, table_out, **tm_kwargs)


class FindCellsTiff(IJTiffFile):
    def __init__(self, fc_kwargs: dict[str, str] = None, *args: Any, **kwargs: Any) -> None:
        self.fc_kwargs = fc_kwargs or {}
        super().__init__(*args, **kwargs)

    def compress_frame(self, frame: tuple[ArrayLike]) -> Sequence[FrameInfo]:
        cell, nucleus = findcells(*frame, **self.fc_kwargs)
        return [super().compress_frame(cell.astype(self.dtype))[0],
                super().compress_frame(nucleus.astype(self.dtype)[0][:2] + ((1, 0, 0),))]


def run_findcells(image: Path, tiff_out: Path, channel_cell: int, channel_nuc: int = None, *,
                  table_out: Path = None, fc_kwargs: dict[str, str] = None, tm_kwargs: dict[str, str] = None) -> None:
    fc_kwargs = fc_kwargs or {}
    tm_kwargs = tm_kwargs or {}
    fc_kwargs = filter_kwargs(findcells, fc_kwargs)

    with tempfile.TemporaryDirectory() as tempdir:
        tif_file = Path(tempdir) / 'tm.tif'

        with Imread(image, axes='ctyx') as im:
            with FindCellsTiff(fc_kwargs, tif_file, (2, 1, im.shape['t']), pxsize=im.pxsize_um) as tif:
                for t in tqdm(range(im.shape['t']), total=im.shape['t'], desc='running findcells',
                              disable=im.shape['t'] < 10):
                    assert channel_cell is not None, 'channel_cell cannot be None'
                    tif.save((im[channel_cell, t],) if channel_nuc is None else
                             (im[channel_cell, t], im[channel_nuc, t]), 0, 0, t)

        trackmate(tif_file, tiff_out, table_out, **tm_kwargs)


class PreTrackTiff(IJTiffFile):
    def __init__(self, shape_yx: tuple[int, int], radius: float, *args: Any, **kwargs: Any) -> None:
        self.shape_yx = shape_yx
        self.xv, self.yv = np.meshgrid(*[range(i) for i in shape_yx])
        self.radius = radius
        super().__init__(*args, **kwargs)

    def compress_frame(self, cxy: tuple[np.ndarray]) -> Sequence[FrameInfo]:
        frame = np.zeros(self.shape_yx, int)
        cxy = cxy[0]
        if len(cxy):
            dist = np.round(np.min([(self.yv - i[1]) ** 2 + (self.xv - i[2]) ** 2 for i in cxy], 0)).astype(int)

            for i in cxy:
                frame[*np.round(i[1:]).astype(int)] = i[0]  # noqa
            frame = watershed(dist, frame, mask=dist < self.radius ** 2)
        return super().compress_frame(frame.astype(self.dtype))


def run_pre_track(image: Path, tiff_out: Path, pre_track: pandas.DataFrame, radius: float) -> None:
    dtype = 'uint8' if pre_track['cell'].max() < 255 else 'uint16'
    with Imread(image) as im:
        with PreTrackTiff(im.shape['yx'], radius, tiff_out, (1, 1, im.shape['t']),  # noqa
                          pxsize=im.pxsize_um, colormap='glasbey', dtype=dtype) as tif:
            for t in tqdm(range(im.shape['t']), total=im.shape['t'], desc='running pre track cell masking',
                          disable=im.shape['t'] < 10):
                tif.save((pre_track.query('T == @t', local_dict=dict(t=t))[['cell', 'y', 'x']].to_numpy(),),
                         0, 0, t)



import re
import xml.etree.cElementTree as et

import numpy as np
import pandas as pd


""" 
    Copyright (c) 2018, Hadrien Mary
    All rights reserved.
    
    Redistribution and use in source and binary forms, with or without
    modification, are permitted provided that the following conditions are
    met:
    
    Redistributions of source code must retain the above copyright notice,
    this list of conditions and the following disclaimer.
    Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the distribution.
    Neither the name of the read_roi nor the names of its
    contributors may be used to endorse or promote products derived from
    this software without specific prior written permission.
    
    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
    "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
    LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
    A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
    HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
    SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
    LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
    DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
    THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
    (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
    OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""


def trackmate_peak_import(trackmate_xml_path, get_tracks=False):
    """Import detected peaks with TrackMate Fiji plugin.

    Parameters
    ----------
    trackmate_xml_path : str
        TrackMate XML file path.
    get_tracks : boolean
        Add tracks to label
    """

    root = et.fromstring(open(trackmate_xml_path).read())

    object_labels = {'FRAME': 't_stamp',
                     'POSITION_T': 't',
                     'POSITION_X': 'x',
                     'POSITION_Y': 'y',
                     'POSITION_Z': 'z',
                     'ESTIMATED_DIAMETER': 'w',
                     'QUALITY': 'q',
                     'ID': 'spot_id',
                     'MEAN_INTENSITY': 'mean_intensity',
                     'MEDIAN_INTENSITY': 'median_intensity',
                     'MIN_INTENSITY': 'min_intensity',
                     'MAX_INTENSITY': 'max_intensity',
                     'TOTAL_INTENSITY': 'total_intensity',
                     'STANDARD_DEVIATION': 'std_intensity',
                     'CONTRAST': 'contrast',
                     'SNR': 'snr'}

    features = root.find('Model').find('FeatureDeclarations').find('SpotFeatures')
    features = [c.get('feature') for c in list(features)] + ['ID']

    spots = root.find('Model').find('AllSpots')
    objects = []
    for frame in spots.findall('SpotsInFrame'):
        for spot in frame.findall('Spot'):
            single_object = []
            for label in features:
                single_object.append(spot.get(label))
            objects.append(single_object)

    trajs = pd.DataFrame(objects, columns=features)
    trajs = trajs.astype(float)

    # Apply initial filtering
    initial_filter = root.find("Settings").find("InitialSpotFilter")

    trajs = filter_spots(trajs,
                         name=initial_filter.get('feature'),
                         value=float(initial_filter.get('value')),
                         isabove=True if initial_filter.get('isabove') == 'true' else False)

    # Apply filters
    spot_filters = root.find("Settings").find("SpotFilterCollection")

    for spot_filter in spot_filters.findall('Filter'):

        trajs = filter_spots(trajs,
                             name=spot_filter.get('feature'),
                             value=float(spot_filter.get('value')),
                             isabove=True if spot_filter.get('isabove') == 'true' else False)

    trajs.columns = [re.sub(r'_CH\d', '', column) for column in trajs.columns]
    trajs.rename(columns=object_labels, inplace=True)
    trajs.columns = [column.lower() for column in trajs.columns]
    trajs['label'] = np.arange(trajs.shape[0])

    # Get tracks
    if get_tracks:
        filtered_track_ids = [int(track.get('TRACK_ID')) for track in root.find('Model').find('FilteredTracks').findall('TrackID')]

        label_id = 0
        trajs['label'] = np.nan

        tracks = root.find('Model').find('AllTracks')
        for track in tracks.findall('Track'):

            track_id = int(track.get("TRACK_ID"))
            if track_id in filtered_track_ids:

                spot_ids = [(edge.get('SPOT_SOURCE_ID'), edge.get('SPOT_TARGET_ID'), edge.get('EDGE_TIME')) for edge in track.findall('Edge')]
                spot_ids = np.array(spot_ids).astype('float')[:, :2]
                spot_ids = set(spot_ids.flatten())

                trajs.loc[trajs["spot_id"].isin(spot_ids), "label"] = label_id
                label_id += 1

        # Label remaining columns
        single_track = trajs.loc[trajs["label"].isnull()]
        trajs.loc[trajs["label"].isnull(), "label"] = label_id + np.arange(0, len(single_track))

    return trajs


def filter_spots(spots, name, value, isabove):
    if isabove:
        spots = spots[spots[name] > value]
    else:
        spots = spots[spots[name] < value]

    return spots
